# Based on the format used in release 119

import Martel
from Martel import RecordReader

header = Martel.Re("""\
(?P<filename>[^ ]+) +Genetic Sequence Data Bank
 *(?P<release_day>\d+) (?P<release_month>\w+) (?P<release_year>\d+)

 *(?P<data_bank_name>[^\R]+)

 *(?P<data_bank_name>[^\R]+)

 *(?P<num_loci>\d+) loci, *(?P<num_bases>\d+) bases, from *(?P<num_reports>\d+) reported sequences


""")

#""" emacs foo

INDENT = 12
FEATURE_KEY_INDENT = 5
FEATURE_QUALIFIER_INDENT = 21

# define yet another special vocabulary

class Vocab:
    def __init__(self, left, right, exp):
        assert left <= right
        self.left = left
        self.right = right
        self.exp = exp
    def __add__(self, other):
        assert isinstance(other, Vocab), other
        assert self.right <= other.left, (self.right, other.left)
        diff = other.left - self.right
        if diff:
            exp = self.exp + Martel.Str(" " * diff) + other.exp
        else:
            exp = self.exp + other.exp
        return Vocab(self.left, other.right, exp)
    
def Word(pos, s):
    return Vocab(pos-1, pos-1+len(s), Martel.Str(s))

def Field(pos, tag, size):
    return Vocab(pos-1, pos-1+size,
                 Martel.Group(tag, Martel.Re("[^\R]" * size)))

def Integer(pos, tag, size, allow_empty = 0):
    # XXX be more precise in the future?
    return Field(pos, tag, size)

def List(pos, tag, words, size, align = "left"):
    assert align in ("left", "right")
    n = max(map(len, words))
    assert n <= size, (n, size)
    n = size
    data = []
    if align == "left":
        for word in words:
            data.append(Martel.Str(word + " " * (n-len(word))))
    else:
        for word in words:
            data.append(Martel.Str(" " * (n-len(word)) + word))
        
    return Vocab(pos-1, pos-1+n, Martel.Group(tag, Martel.Alt(*data)))
    

def Date(pos, tag):
    return Vocab(pos-1, pos-1+11,
                 Martel.Group(tag+"_date",
                      Martel.Group(tag+"_day", Martel.Re("\d\d")) + \
                      Martel.Str("-") + \
                      Martel.Group(tag+"_month", Martel.Re("\w\w\w")) + \
                      Martel.Str("-") + \
                      Martel.Group(tag+"_year", Martel.Re("\d\d\d\d"))))

def convert(vocab):
    assert vocab.left >= 0
    if vocab.left != 0:
        exp = Martel.Str(" " * vocab.left) + vocab.exp
    else:
        exp = vocab.exp
    return exp + Martel.AnyEol()

def define_block(identifier, block_tag, block_data):
    diff = INDENT - len(identifier)
    assert diff > 0, diff

    return Martel.Group(block_tag,
                        Martel.Str(identifier + " " * diff) + \
                        Martel.ToEol(block_data) + \
                        Martel.Rep(Martel.Str(" " * INDENT) + \
                                   Martel.ToEol(block_data)))

def define_multiple(identifier, block_tag, block_data, exp):
    diff = INDENT - len(identifier)
    assert diff > 0, diff
    data = Martel.Group(block_data,
                        exp + \
                        Martel.Rep(Martel.Str(" ") + exp) + \
                        Martel.AnyEol())
                        
    return Martel.Group(block_tag,
                        Martel.Str(identifier + " " * diff) + \
                        data + \
                        Martel.Rep(Martel.Str(" " * INDENT) + \
                                   data))


locus = convert(Word(1, "LOCUS") + \
                Field(13, "locus_name", 10) + \
                Integer(23, "length", 7) + \
                Word(31, "bp") + \
                List(34, "strands", ["", "ss-", "ds-", "ms-"], 3) + \
                List(37, "topology", ["", "DNA", "RNA", "tRNA", "rRNA",
                                      "mRNA", "uRNA"], 4) + \
                List(43, "special_sequence", ["", "circular"], 10) + \
                Field(53, "division", 3) + \
                Date(63, "entry")
                )


definition = define_block("DEFINITION", "definition_block", "definition")

accession_number = Martel.Group("accession_number",
                                Martel.Re(r"[A-Z]\d{5}|[A-Z]{2}\d{6}"))

accession = define_multiple("ACCESSION", "accession_block", "accession",
                            accession_number)

gi_identifier = Martel.Group("gi_identifier",
                             Martel.Re(r"GI:\d+"))

version = Martel.Group("version_block",
                       Martel.Str("VERSION     ") + \
                       Martel.Group("compound_accession_number",
                                    accession_number + \
                                    Martel.Str(".") + \
                                    Martel.Integer("accession_version")
                                    ) + \
                       Martel.Str("  ") + \
                       gi_identifier + \
                       Martel.AnyEol())


keywords = define_block("KEYWORDS", "keywords_block", "keywords")

segment = Martel.Re(r"SEGMENT     (?P<segment_number>\d+) of (?P<num_segments>\d+)\R")

source = define_block("SOURCE", "source_block", "source")

# I know the genus/species is wrong.  A real definition needs a lot
# of work to parse the somewhat unconstrained text.
organism = Martel.Re(r"  ORGANISM  ") + \
           Martel.Group("scientific_name",
                        Martel.Re(r"(?P<genus>\w+) (?P<species>\w+)(?=\\R)") |
                        Martel.Re(r"[^\R]*")) + \
           Martel.AnyEol() + \
           define_block("", "taxonomy_block", "taxonomy")

reference_line = Martel.Re(r"REFERENCE   (?P<reference_number>\d+)") + \
                 Martel.Group("bases_range_block", 
                              Martel.ToEol("bases_range") + \
                              Martel.Rep(Martel.Str(" " * INDENT) +
                                         Martel.ToEol("bases_range")))

authors = define_block("  AUTHORS", "authors_block", "authors")
title = define_block("  TITLE", "title_block", "title")
journal = define_block("  JOURNAL", "journal_block", "journal")
medline = Martel.Str("  MEDLINE   ") + \
          Martel.Integer("medline_ui") + \
          Martel.AnyEol()

pubmed = Martel.Str("   PUBMED   ") + \
         Martel.Integer("pubmed_ui") + \
         Martel.AnyEol()

remark = define_block("  REMARK", "remark_record", "remark")

reference = Martel.Group("reference",
                         reference_line + \
                         authors + \
                         Martel.Opt(title) + \
                         journal + \
                         Martel.Opt(medline) + \
                         Martel.Opt(pubmed) + \
                         Martel.Opt(remark))

comment = define_block("COMMENT", "comment_block", "comment")

# XXX Because of a bug in Martel, if one name is a substring of
# XXX another, put the longer name first.  Eg, "primer_bind" before "bind"
feature_key_names = (
    "allele",           # Obsolete; see variation feature key
    "attenuator",       # Sequence related to transcription termination
    "C_region",         # Span of the C immunological feature
    "CAAT_signal",      # 'CAAT box' in eukaryotic promoters
    "CDS",              # Sequence coding for amino acids in protein (includes
                        #   stop codon)
    "conflict",         # Independent sequence determinations differ
    "D-loop",           # Displacement loop
    "D_segment",        # Span of the D immunological feature
    "enhancer",         # Cis-acting enhancer of promoter function
    "exon",             # Region that codes for part of spliced mRNA
    "GC_signal",        # 'GC box' in eukaryotic promoters
    "gene",             # Region that defines a functional gene, possibly
                        #   including upstream (promotor, enhancer, etc)
                        #   and downstream control elements, and for which
                        #   a name has been assigned.
    "iDNA",             # Intervening DNA eliminated by recombination
    "intron",           # Transcribed region excised by mRNA splicing
    "J_segment",        # Span of the J immunological feature
                        # (relnotes says 'J_region' but J_segment in files)
    "LTR",              # Long terminal repeat
    "mat_peptide",      # Mature peptide coding region (does not include
                        #   stop codon)
    "misc_binding",     # Miscellaneous binding site
    "misc_difference",  # Miscellaneous difference feature
    "misc_feature",     # Region of biological significance that cannot
                        #   be described by any other feature
    "misc_recomb",      # Miscellaneous recombination feature
    "misc_RNA",         # Miscellaneous transcript feature not defined by
                        #   other RNA keys
    "misc_signal",      # Miscellaneous signal
    "misc_structure",   # Miscellaneous DNA or RNA structure
    "modified_base",    # The indicated base is a modified nucleotide
    "mRNA",             # Messenger RNA
    "mutation",         # Obsolete: see variation feature key
    "N_region",         # Span of the N immunological feature
    "old_sequence",     # Presented sequence revises a previous version
    "polyA_signal",     # Signal for cleavage & polyadenylation
    "polyA_site",       # Site at which polyadenine is added to mRNA
    "precursor_RNA",    # Any RNA species that is not yet the mature
                        #   RNA product
    "prim_transcript",  # Primary (unprocessed) transcript
    "primer_bind",      # Non-covalent primer binding site
    "primer",           # Primer binding region used with PCR  XXX not in 
                        #   http://www.ncbi.nlm.nih.gov/collab/FT/index.html
    "promoter",         # A region involved in transcription initiation
    "protein_bind",     # Non-covalent protein binding site on DNA or RNA
    "RBS",              # Ribosome binding site
    "rep_origin",       # Replication origin for duplex DNA
    "repeat_region",    # Sequence containing repeated subsequences
    "repeat_unit",      # One repeated unit of a repeat_region
    "rRNA",             # Ribosomal RNA
    "S_region",         # Span of the S immunological feature
    "satellite",        # Satellite repeated sequence
    "scRNA",            # Small cytoplasmic RNA
    "sig_peptide",      # Signal peptide coding region
    "snRNA",            # Small nuclear RNA
    "source",           # Biological source of the sequence data
                        #   represented by a GenBank record. Mandatory
                        #   feature, one or more per record.  For organisms
                        #   that have been incorporated within the NCBI
                        #   taxonomy database, an associated
                        #   /db_xref="taxon:NNNN" qualifier will be present
                        #   (where NNNNN is the numeric identifier assigned
                        #   to the organism within the NCBI taxonomy
                        #   database).
    "stem_loop",        # Hair-pin loop structure in DNA or RNA
    "STS",              # Sequence Tagged Site; operationally unique
                        #   sequence that identifies the combination of
                        #   primer spans used in a PCR assay
    "TATA_signal",      # 'TATA box' in eukaryotic promoters
    "terminator",       # Sequence causing transcription termination
    "transit_peptide",  # Transit peptide coding region
    "transposon",       # Transposable element (TN)
    "tRNA",             # Transfer RNA
    "unsure",           # Authors are unsure about the sequence in this region
    "V_region",         # Span of the V immunological feature
    "V_segment",        # Variable segment of immunoglobulin light and heavy
                        #   chains, and T-cell receptor alpha, beta, and
                        #   gamma chains
    "variation",        # A related population contains stable mutation
    "-",                # (hyphen)      Placeholder
    "-10_signal",       # 'Pribnow box' in prokaryotic promoters
    "-35_signal",       # '-35 box' in prokaryotic promoters
    "3'clip",           # 3'-most region of a precursor transcript removed'
                        #   in processing
    "3'UTR",            # 3' untranslated region (trailer)'
    "5'clip",           # 5'-most region of a precursor transcript removed'
                        # in processing
    "5'UTR",            # 5' untranslated region (leader)'
)

# For now use a simple list and get all of the matching text.
# In the future could allow more specific matches for each key.
# XXX Because of a bug in Martel, if one name is a substring of
# XXX another, put the longer name first.  Eg, "clone_lib" before "clone"
feature_qualifier_names = (
    "allele",         # Name of the allele for the a given gene
    "anticodon",      # Location of the anticodon of tRNA and the amino
                      #   acid for which it codes
    "bound_moiety",   # Moiety bound
    "cell_line",      # Cell line from which the sequence was obtained
    "cell_type",      # Cell type from which the sequence was obtained
    "chromosome",     # Chromosome (e.g. Chromosome number) from which
                      #   the sequence was obtained
    "chloroplast",    # Organelle type from which the sequence was obtained
    "chromoplast",    # Organelle type from which the sequence was obtained
    "citation",       # Reference to a citation providing the claim of or
                      #   evidence for a feature
    "clone_lib",      # Clone library from which the sequence was obtained
    "clone",          # Clone from which the sequence was obtained
    "codon_start",    # Indicates the first base of the first complete codon
                      #   in a CDS (as 1 or 2 or 3)
    "codon",          # Specifies a codon that is different from any found
                      #   in the reference genetic code
    "cons_splice",    # Identifies intron splice sites that do not conform to
                      #   the 5'-GT... AG-3' splice site consensus
    "country",        # Country of origin for DNA sample, intended for
                      #   epidemiological or population studies.
    "cultivar",       # Variety of plant from which sequence was obtained
    "cyanelle",       # Organelle type from which the sequence was obtained
    "db_xref",        # A database cross-reference; pointer to related
                      #   information in another database. A description of
                      #   all cross-references can be found at:
                      #   http://www.ncbi.nlm.nih.gov/collab/db_xref.html
    "dev_stage",      # If the sequence was obtained from an organism in
                      #   a specific developmental stage, it is specified
                      #   with this qualifier
    "direction",      # Direction of DNA replication
    "EC_number",      # Enzyme Commission number for the enzyme product
                      #   of the sequence
    "evidence",       # Value indicating the nature of supporting evidence
    "exception",      # Indicates that the amino acid or RNA sequence
                      #   will not translate or agree with the DNA sequence
                      #   according to standard biological rules
    "focus",          # Defines the source feature of primary biological
                      #   interest for records that have multiple source
                      #   features originating from different organisms
    "frequency",      # Frequency of the occurrence of a feature
    "function",       # Function attributed to a sequence
    "gene",           # Symbol of the gene corresponding to a sequence
                      #   region (usable with all features)
    "germline",       # If the sequence shown is DNA and a member of the
                      #   immunoglobulin family, this qualifier is used to
                      #   denote that the sequence is from unrearranged DNA
    "haplotype",      # Haplotype of organism from which the sequence was
                      #   obtained
    "insertion_seq",  # Insertion sequence element from which the sequence
                      #   was obtained
    "isolate",        # Individual isolate from which the sequence was
                      #   obtained
    "kinetoplast",    # Organelle type from which the sequence was obtained
    "label",          # A label used to permanently identify a feature
    "lab_host",       # Laboratory host used to propagate the organism
                      #   from which the sequence was obtained
    "macronuclear",   # If the sequence shown is DNA and from an organism
                      #   which undergoes chromosomal differentiation
                      #   between macronuclear and micronuclear stages,
                      #   this qualifier is used to denote that the
                      #   sequence is from macronuclear DNA.
    "map",            # Map position of the feature in free-format text
    "mitochondrion",  # Organelle type from which the sequence was obtained
    "mod_base",       # Abbreviation for a modified nucleotide base
    "note",           # Any comment or additional information
    "number",         # A number indicating the order of genetic elements
                      #   (e.g., exons or introns) in the 5 to 3 direction
    "organelle",      # Type of membrane-bound intracellular structure from
                      #  which the sequence was obtained
    "organism",       # Name of the organism that is the source of the
                      #   sequence data in the record. 
    "partial",        # Differentiates between complete regions and
                      #   partial ones
    "PCR_conditions", # Description of reaction conditions and components
                      #   for PCR
    "phenotype",      # Phenotype conferred by the feature
    "plasmid",        # Name of plasmid from which sequence was obtained
    "pop_variant",    # Population variant from which the sequence was obtained
    "product",        # Name of a product encoded by a coding region (CDS)
                      #   feature
    "protein_id",     # Protein Identifier, issued by International
                      #   collaborators.  This qualifier consists of a stable
                      #   ID portion (3+5 format with 3 position letters and
                      #   5 numbers) plus a version number after the decimal
                      #   point
    "proviral",       # If the sequence shown is viral and integrated into
                      #   another organism's genome, this qualifier is used
                      #   to denote that.
    "pseudo",         # Indicates that this feature is a non-functional
                      #   version of the element named by the feature key
    "rearranged",     # If the sequence shown is DNA and a member of the
                      #   immunoglobulin family, this qualifier is used to
                      #   denote that the sequence is from rearranged DNA
    "replace",        # Indicates that the sequence identified a feature's
                      #   intervals is replaced by the  sequence shown in
                      #   "text"
    "rpt_family",     # Type of repeated sequence; Alu or Kpn, for example
    "rpt_type",       # Organization of repeated sequence
    "rpt_unit",       # Identity of repeat unit that constitutes a
                      #   repeat_region
    "sequenced_mol",  # Molecule from which the sequence was obtained
    "serotype",       # Variety of a species (usually bacteria or virus)
                      #   characterized by its antigenic properties
    "sex",            # Sex of the organism from which the sequence
                      #   was obtained
    "specific_host",  # Natural host from which the sequence was obtained
    "specimen_voucher", # An identifier of the individual or collection
                      #   of the source organism and the place where it
                      #   is currently stored, usually an institution.
    "standard_name",  # Accepted standard name for this feature
    "strain",         # Strain from which the sequence was obtained
    "sub_clone",      # Sub-clone from which the sequence was obtained
    "sub_species",    # Sub-species name of organism  from which the
                      #   sequence was obtained
    "sub_strain",     # Sub_strain from which the sequence was obtained
    "tissue_lib",     # Tissue library from which the sequence was obtained
    "tissue_type",    # Tissue type from which the sequence was obtained
    "translation",    # Amino acid translation of a coding region
    "transl_except",  # Translational exception: single codon, the
                      #   translation of which does not conform to the
                      #   reference genetic code
    "transl_table",   # Definition of genetic code table used if other
                      #   than universal genetic code table
    "transposon",     # Transposable element from which the sequence
                      #   was obtained
    "type",           # Name of a strain if different from that in the
                      #   SOURCE field  (XXX not in
                      #   http://www.ncbi.nlm.nih.gov/collab/FT/index.html )
    "usedin",         # Indicates that feature is used in a compound
                      #   feature in another entry
    "variety",        # Variety from which sequence was obtained
    "virion",         # Viral genomic sequence as it is encapsidated
                      #   (distinguished from its proviral form integrated
                      #   in a host cell's chromosome) 
    
)

features_line = Martel.Str("FEATURES             Location/Qualifiers") + \
                Martel.AnyEol()

# From the start of the feature key to the beginning of the qualifer column
feature_keys = [Martel.Group("feature_key", Martel.Str(x)) + \
                Martel.Str(" " * (FEATURE_QUALIFIER_INDENT - \
                                  FEATURE_KEY_INDENT - len(x)))
                for x in feature_key_names]
feature_keys = Martel.Alt(*feature_keys)


feature_location = Martel.Group("feature_location_block",
                      Martel.ToEol("feature_location") + \
                      Martel.Rep(Martel.Str(" " * FEATURE_QUALIFIER_INDENT) + \
                                 Martel.Re("(?!/)") + \
                                 Martel.ToEol("feature_location")))

qualifier_names = map(Martel.Str, feature_qualifier_names)
qualifier_names = Martel.Group("qualifier_name", Martel.Alt(*qualifier_names))

qualifier_block = Martel.Group("qualifier_block",
                               Martel.Str(" " * FEATURE_QUALIFIER_INDENT + \
                                          "/") + \
                               qualifier_names + \
                               (Martel.Re(r"\R") | \
                                Martel.Str("=") + \
                                Martel.ToEol("qualifier_data") + \
                                Martel.Rep(
                         Martel.Str(" " * FEATURE_QUALIFIER_INDENT) + \
                         Martel.Re("(?!/)") + \
                         Martel.ToEol("qualifier_data"))))

feature_key_location = Martel.Str(" " * FEATURE_KEY_INDENT) + \
                       feature_keys + \
                       feature_location

feature_key_block = Martel.Group("feature_key_block",
                                 feature_key_location + \
                                 Martel.Rep(qualifier_block))

features = features_line + \
           Martel.Rep1(feature_key_block)


base_count = Martel.Re(r"BASE COUNT +(?P<num_a>\d+) a +(?P<num_c>\d+) c +(?P<num_g>\d+) g +(?P<num_t>\d+) t( *(?P<num_other>\d+) others)?\R")

# All the records have the spaces after the ORIGIN, but I don't
# trust people quite that much, so let it be optional.
origin_line = Martel.Str("ORIGIN") + \
              ((Martel.Str("      ") + Martel.ToEol("origin_data")) |
               Martel.AnyEol())

sequence = Martel.Re("(?P<position>.........) ") + \
           Martel.ToEol("sequence")

sequence_block = Martel.Group("sequence_block", Martel.Rep(sequence))
                     
end_of_record = Martel.Str("//") + Martel.AnyEol()

record = Martel.Group("genbank_record",
                      Martel.Re(r"\R?") + \
                      locus + \
                      definition + \
                      accession + \
                      Martel.Opt(version) + \
                      keywords + \
                      Martel.Opt(segment) + \
                      source + \
                      organism + \
                      Martel.Rep(reference) + \
                      Martel.Opt(comment) + \
                      features + \
                      base_count + \
                      origin_line + \
                      sequence_block + \
                      end_of_record)

format = Martel.HeaderFooter("genbank",
                             header, RecordReader.CountLines, (10,),
                             record, RecordReader.EndsWith, ("//",),
                             None, None, None,
                             )

multirecord = Martel.ParseRecords("genbank", record,
                                  RecordReader.EndsWith, ("//",))
