# Parse the CODATA format.  Tested against 
# ftp://nbrfa.georgetown.edu/pir/databases/pir_codata/

# Based on (incomplete) format documentation for 3.0 (dated "December
# 31, 1993") found at
# ftp://nbrfa.georgetown.edu/pir/documents/0protein_doc.codata


from Martel import Re, Str, Rep, RepN, Group, Opt, Integer, Rep1
from Martel import HeaderFooter, ToEol, Alt, AnyEol

from Martel import RecordReader
from Martel.Expression import Assert

header0 = Opt(Re(r" *\R"))

header1 = Str("""\
                P R O T E I N  S E Q U E N C E  D A T A B A S E
                             of PIR-International
 
""")
header2 = Re("""\
 *Section (?P<section_id>\d+). (?P<section_title>[^\R]*)
 *Release (?P<release>(?P<major_version>\d+)\.(?P<minor_version>\d+)), *(?P<date>[^\R]*)
 *(?P<num_sequences>\d+) sequences, (?P<num_residues>\d+) residues
 
""")
header3 = Re("""\
 *Protein Information Resource \(PIR\)\*
 *National Biomedical Research Foundation
 *3900 Reservoir Road, N\.W\.,
 *Washington, DC  20007, USA
 *
 *Japan International Protein *Munich Information Center for
 *Information Database \(JIPID\) *Protein Sequences \(MIPS\)
 *Amakubo 1-16-1 *GSF-Forschungszentrum f\. Umwelt und Gesundheit
 *Tsukuba 305-0005, Japan *am Max-Planck-Instut f\. Biochemie
 *Am Klopferspitz 18, D-82152 Martinsried, FRG
 *
""")
header4 = Re("""\
 *This database may be redistributed without prior consent, provided that
 *this notice be given to each user and that the words "Derived from" shall
 *precede this notice if the database has been altered by the redistributor\.
 *
((?!    *Copyright) [^\R]*
)* *Copyright (?P<copyright_year>\d+), PIR-International\.
 *
 *\*PIR is a registered mark of NBRF\.
( *
)?\\\\\\\\\\\\
""")
header =  Group("header", header0 + header1 + header2 + header3 + header4)
    

""" \"\"" """  # EMACS doesn't like all the quoting above, so fake it out

# Most lines are of the form
# <spaces>* <label> <spaces>+ data
#
# If the data folds over it is broken at a word break and indented by
# N number of spaces (N depends on where the data is in the file)

SUBITEM_INDENT = 3
SUBSUBITEM_INDENT = 6
DATA_INDENT = 16
FEATURE_INDENT = 23

def make_wrap(indent):
    return AnyEol() + Str(" " * indent)

class Data:
    def __init__(self, tag, can_skip = 0):
        self.tag = tag
        self.can_skip = can_skip
    def emit(self, indent):
        raise NotImplementedError

# Read up to the spaces before the '#' or up to just before the newline
class Text(Data):
    def emit(self, indent):
        wrap = make_wrap(indent)
        words = Group(self.tag, Re("([^ #\R][^ \R]*( +(?!#))?)+"))
        return words + Rep(wrap + words)

class Word(Data):
    def emit(self, indent):
        return Group(self.tag, Re("[^ \R]+"))
  
class NamedText(Data):
    def __init__(self, tag, name, allow_empty = 0, can_skip = 0):
        Data.__init__(self, tag, can_skip)
        self.name = name
        self.allow_empty = allow_empty
        
    def emit(self, indent):
        text = Text(self.tag).emit(indent)
        wrap = make_wrap(indent)
        if not self.allow_empty:
            return Str(self.name) + ((Str(" ") + text) | (wrap + text))
        else:
            # Using the assertion forces a begin/end tag to be generated
            return Str(self.name) + (((Str(" ") + text) | (wrap + text)) | \
                                     Group(self.tag, Re(r"(?=.|\R)")))
    
                   
class OptionList(Data):
    # XXX Possible bug if there's a space in the option which need to wrap.
    # Can be fixed.  See if it's needed.
    def __init__(self, tag, name, opts, can_skip = 0):
        Data.__init__(self, tag, can_skip)
        self.name = name
        self.opts = opts
    def emit(self, indent):
        wrap = make_wrap(indent)
        exp = Group(self.tag, apply(Alt, map(Str, self.opts)))
        if self.name is not None:
            return Str(self.name) + \
                   ((Str(" ") + exp) | (wrap + exp))
                    
        else:
            return (Str(" ") + exp) | (wrap + exp)

class OneOf(Data):
    def __init__(self, *opts):
        Data.__init__(self, "<no tag>", 0)
        self.opts = opts
    def emit(self, indent):
        exp = apply(Alt, [x.emit(indent) for x in self.opts])
        return exp

def _make_fields(spacer, need_spacer, indent, fields):
    field = fields[0]
    if field.can_skip:
        if need_spacer:
            x = spacer + field.emit(indent)
        else:
            x = field.emit(indent)
        if fields[1:]:
            exp = ( x + _make_fields(spacer, 1, indent, fields[1:]) ) | \
                  _make_fields(spacer, need_spacer, indent, fields[1:])
        else:
            exp = Opt(x)
    else:
        if need_spacer:
            x = spacer + field.emit(indent)
        else:
            x = field.emit(indent)
        if fields[1:]:
            exp = x + _make_fields(spacer, 1, indent, fields[1:])
        else:
            exp = x
    return exp

def make_generic_data(prefix, fields, indent = DATA_INDENT):
    assert len(fields) > 0
    wrap = make_wrap(indent)
    spacer = Rep1(Str(" ")) | wrap

    field = fields[0]
    if field.can_skip:
        if fields[:1]:
            exp = (field.emit(indent) + \
                   _make_fields(spacer, 1, indent, fields[1:])) | \
                   _make_fields(spacer, 0, indent, fields[1:])
        else:
            exp = Opt(field.emit(indent))
    else:
        if fields[1:]:
            exp = field.emit(indent) + \
                  _make_fields(spacer, 1, indent, fields[1:])
        else:
            exp = field.emit(indent)

    return prefix + exp + AnyEol()

def make_identifier(label, fields, allow_newline = 0):
    if not fields:
        return Str(label) + AnyEol()
    count = max(1, DATA_INDENT - len(label))
    exp = Str(label + " " * count)
    if allow_newline:
        # It's weird, but a few records have it
        exp = exp | (Str(label + " " * (count-1)) + AnyEol() + \
                     Str(" " * DATA_INDENT))
    return make_generic_data(exp, fields)

def make_subitem(label, fields):
    count = max(1, DATA_INDENT - SUBITEM_INDENT - len(label))
    exp = Str(" " * SUBITEM_INDENT + label + " " * count)
    return make_generic_data(exp, fields)

def make_subsubitem(label, fields):
    count = max(1, DATA_INDENT - SUBSUBITEM_INDENT - len(label))
    exp = Str(" " * SUBSUBITEM_INDENT + label + " " * count)
    return make_generic_data(exp, fields)
    

#ENTRY           CCHU  #type complete
entry = make_identifier("ENTRY",
                        (Word("entry_id"),
                         OptionList("entry_type", "#type",
                                    ("complete", "fragments", "fragment"))))

#F69392 has a TITLE which starts on the second line, and H82063
title = make_identifier("TITLE",
                        (Text("title"),),
                        allow_newline = 1,
                       )

alternate_names = make_identifier("ALTERNATE_NAMES",
                                  (Text("alternate_names"),),
                                  allow_newline = 1,
                                 )

contains = make_identifier("CONTAINS", (Text("contains"),) )

organism_identifier = make_identifier("ORGANISM",
                           (NamedText("formal_name", "#formal_name"),
                            NamedText("common_name", "#common_name",
                                      can_skip = 1),
                            )
                           )
organism_variety = make_subitem("#variety", (Text("variety"), ))
organism_note = make_subitem("#note", (Text("note"), ))

organism = organism_identifier + Opt(organism_variety) + Opt(organism_note)


date = make_identifier("DATE",
                       (Text("date"),
                        NamedText("sequence_revision", "#sequence_revision"),
                        NamedText("text_change", "#text_change"),
                        )
                       )

accessions = make_identifier("ACCESSIONS",
                             (Text("accessions"),)
                             )
                             
ref_number = make_identifier("REFERENCE",
                             (Text("reference_number"),)
                             )

ref_authors = make_subitem("#authors",
                           (Text("reference_authors"),)
                           )

# Can go over two lines!
journal_subitem = Group("journal_subitem",
                        Str("   #journal     ") + \
                        Group("journal_abbreviation", Re(r"[^(\R]+")) + \
                        Re(r"\((?P<year>\d+)\)") + \
                        Opt(AnyEol() + Str("               ")) + \
                        Str(" ") + \
                        Opt(Re(r"(?P<volume>[^[:\R]+)?")) + \
                        Opt(Re(r"(\[(?P<issue>\d+)\])?")) + \
                        Str(":") + \
                        Group("pages",
                              Re(r"(?P<start_page>\d+)(-(?P<end_page>\d+))?")) + \
                        AnyEol())

#need to handle
#journal     Mol. Biol. (Mosk.) (1990) 24:968-976
journal_subitem = make_subitem("#journal", (Text("journal"),))
                        
ref_submission = make_subitem("#submission", (Text("submission"),))

# Need examples for these
ref_book = make_subitem("#book", (Text("book"),))
ref_citation = make_subitem("#citation", (Text("citation"),))

title_subitem = Group("title_subitem",
                      Str("   #title       ") + \
                      ToEol("title") + \
                      Rep(Str("                ") + \
                          ToEol("title"))
                      )

ref_description = make_subitem("#description",
                               (Text("reference_description"),))

ref_title_subitem = Group("reference_title_subitem",
                          title_subitem | ref_description)

# cross-references to medline or PDB
xref_subitem = Group("cross_reference_subitem",
                     Str("   #cross-references ") + \
                     Group("xref",
                           Group("xref_database", Re("MUID|PDB|BMRB")) + \
                           Str(":") + \
                           Group("xref_id", Re(r"[^\R]+"))) + \
                     AnyEol())

contents_subitem = Str("   ") + \
                    Group("contents_subitem",
                          Str("#contents    ") + \
                          ToEol("contents_data") + \
                          Rep(AnyEol() + Str("                ") + \
                              Group("contents_data", Re(r"[^\R]*"))) + \
                          AnyEol())

note_subitem = Str("   ") + \
               Group("note_subitem",
                     Str("#note        ") + \
                     Group("note_data", Re(r"[^\R]*")) + \
                     Rep(AnyEol() + Str("                ") + \
                         Group("note_data", Re(r"[^\R]*"))) + \
                     AnyEol())

ref_acc = make_subitem("#accession",
                       (Text("reference_accession"),))

# The first line isn't indented to the DATA_INDENT level
ref_acc_status = make_generic_data(Str("      ##status "),
                                   (Text("reference_accession_status"),))

ref_acc_molecule_type = make_subsubitem("##molecule_type",
                                        (Text("molecule_type"),))

# This can have an extra newline+indent inserted
ref_acc_residues = make_generic_data(Str("      ##residues") + \
                                     Alt(Str(" ")|(AnyEol() + Str(" " * 16))),
                                     (Text("residues_range"),
                                      NamedText("residues_label",
                                                "##label")))
# The id can be "", at least, so says T50374 in pir3 which has an
# "SPDB:"
ref_acc_xrefs = make_subsubitem("##cross-references",
                                (Text("reference_accession_xrefs"),))

# G72850 has the #date and #staff
ref_acc_experimental_source = make_subsubitem("##experimental_source",
                      (Text("reference_experimental_source"),
                       NamedText("annotate_date", "#date", can_skip=1),
                       NamedText("annotate_staff", "#staff", can_skip=1),
                                ))

ref_acc_genetics = make_subsubitem("##genetics",
                                   (Text("reference_accession_genetics"),))

# The first line isn't indented to the DATA_INTENT level
# A45121 has the #date and #staff
ref_acc_note = make_generic_data(Str("      ##note "),
                         (Text("reference_accession_note"),
                          NamedText("annotate_date", "#date", can_skip=1),
                          NamedText("annotate_staff", "#staff", can_skip=1),
                                 ))

# DEKCNG, C24706 don't have a ref_acc_residues
ref_acc_subitem = Group("reference_accession_subitem",
                        ref_acc + \
                        Opt(ref_acc_status) + \
                        Opt(ref_acc_molecule_type) + \
                        Opt(ref_acc_residues) + \
                        Opt(ref_acc_xrefs) + \
                        Opt(ref_acc_experimental_source) + \
                        Opt(ref_acc_genetics) + \
                        Rep(ref_acc_note)
                        )
                          
# not all journals have titles
# KKGT has a #citation followed by #title 
reference = ref_number + \
            ref_authors + \
            ((journal_subitem + Opt(title_subitem)) | \
             (ref_submission + Opt(ref_description)) | \
             (ref_book + Opt(title_subitem)) | \
             (ref_citation + Opt(ref_description|title_subitem))) + \
            Opt(xref_subitem) + \
            Opt(contents_subitem) + \
            Rep(note_subitem) + \
            Rep(ref_acc_subitem)

comment = make_identifier("COMMENT", (Text("comment"),))
            

genetics_identifier = (Str("GENETICS") + AnyEol()) | \
                      make_identifier("GENETICS",
                                      (Word("genetics_identifier"),))
genetics_gene = make_subitem("#gene", (Text("genetics_gene"),))
genetics_gene_xrefs = make_subsubitem("##cross-references",
                                      (Text("genetics_gene_xrefs"),))
genetics_map_position = make_subitem("#map_position",
                                     (Text("genetics_map_position"),))
genetics_genome = make_subitem("#genome", (Text("genetics_genome"),))
genetics_genetic_code = make_subitem("#genetic_code",
                                     (Text("genetics_genetic_code"),))
genetics_start_codon = make_subitem("#start_codon",
                                    (Text("genetics_start_codon"),))
                                               
genetics_introns = make_subitem("#introns",
                       (Text("genetics_introns", can_skip = 1),
                        NamedText("introns_status", "#status", can_skip = 1))
                                )
                                 
genetics_mobile_element = make_subitem("#mobile_element",
                                       (Text("genetics_mobile_element"),))
genetics_note = make_subitem("#note", (Text("genetics_note"),))

# XXX a few records reverse the start_codon and mobile_element field
# I fake it by allowing two start_codon fields
# T02916,T10803,others allow #introns after mobile_element
genetics = genetics_identifier + \
           Opt(genetics_gene) + \
           Opt(genetics_gene_xrefs) + \
           Opt(genetics_map_position) + \
           Opt(genetics_genome) + \
           Opt(genetics_genetic_code) + \
           Opt(genetics_start_codon) + \
           Opt(genetics_introns) + \
           Opt(genetics_mobile_element + \
               Opt(genetics_start_codon) + \
               Opt(genetics_introns) + \
               Opt(genetics_genetic_code)) + \
           Rep(genetics_note)



complex = make_identifier("COMPLEX", (Text("complex"),))


# was Word instead of Text
function_identifier = (Str("FUNCTION") + AnyEol()) | \
                      make_identifier("FUNCTION",
                                      (Text("function_identifier"),))
function_description = make_subitem("#description",
                                    (Text("function_description"),))
function_pathway = make_subitem("#pathway", (Text("function_pathway"),))
function_note = make_subitem("#note", (Text("function_node"),))

# description not given in B42147, S13744, A47620
function = function_identifier + \
           Opt(function_description) + \
           Opt(function_pathway) + \
           Rep(function_note)


classification = make_identifier("CLASSIFICATION",
                                 (NamedText("classification_superfamily",
                                            "#superfamily"),))

keywords = make_identifier("KEYWORDS", (Text("keywords"),))

# fragment definition
feature_pos = Re(r"(?P<feature_pos>\d+)")
feature_seq = Re(r"'(?P<feature_seq>[A-Z]+)'")
feature_range = Re(r"(?P<feature_range>(?P<begin_pos>\d+)-(?P<end_pos>\d+))")
feature_element = feature_range | feature_pos | feature_seq

fragment = feature_element + \
           Rep(Str(",") + \
               Opt(AnyEol() + Str(" " * SUBITEM_INDENT)) + \
               feature_element)

feature_identifier = make_identifier("FEATURE", ())
# cross_link
feature_data = make_generic_data( Str(" " * SUBITEM_INDENT) + \
                                  fragment + Re(" *"),
             (OneOf(
    NamedText("active_site", "#active_site"),
    NamedText("binding_site", "#binding_site"),
    NamedText("cleavage_site", "#cleavage_site"),
    NamedText("inhibitory_site", "#inhibitory_site"),
    NamedText("modified_site", "#modified_site"),
    NamedText("product", "#product", allow_empty = 1),  # T07826, A55362
    NamedText("disulfide_bonds", "#disulfide_bonds", allow_empty = 1),
    NamedText("cross_links", "#cross-links"),
    NamedText("domain", "#domain"),
    NamedText("region", "#region"),
    ),
              NamedText("link", "#link", can_skip = 1),
              NamedText("status", "#status", can_skip = 1),
              NamedText("label", "#label", can_skip = 1),
              ),
                                  indent = FEATURE_INDENT)

feature = feature_identifier + \
          Rep1(Group("feature", feature_data))

summary = make_identifier("SUMMARY",
                          (NamedText("summary_length", "#length"),
                           NamedText("summary_mw", "#molecular-weight",
                                     can_skip = 1),
                           NamedText("summary_checksum", "#checksum")))

sequence_identifier = make_identifier("SEQUENCE", ())
sequence_numbers = Re(r"        [^\R]*\R")

sequence_punc = Re(" |(?P<punctuation>[()=/.,])")
sequence_letter = Re("(?P<sequence>[A-Z])")
sequence_punc_letter = sequence_punc + sequence_letter
# The hack is to work around a Martel bug
sequence_data = Rep1(Re("   (?P<start_pos>.....)") + \
                     Rep(Group("hack", sequence_punc_letter)) + \
                     Opt(sequence_punc) + AnyEol())

sequence = sequence_identifier + \
           sequence_numbers + \
           Rep(Group("sequence_data", sequence_data))

end_of_entry = Str("///") + AnyEol()


record = entry + \
         title + \
         Opt(alternate_names) + \
         Opt(contains) + \
         organism + \
         date + \
         accessions + \
         Rep1(reference) + \
         Rep(comment) + \
         Rep(genetics) + \
         Rep(complex) + \
         Rep(function) + \
         Opt(classification) + \
         Opt(keywords) + \
         Opt(feature) + \
         summary + \
         sequence + \
         end_of_entry

footer = Str("\\\\\\") + AnyEol()

format = HeaderFooter("pir3_0",
                      header, RecordReader.Until, ("ENTRY", ),
                      record, RecordReader.EndsWith, ("///", ),
                      footer, RecordReader.StartsWith, ("\\\\\\", ),)
                      
