"""

Parses various formats from 2.0.10

Only been tested with the biopython files biopython/Test/Blast/

bt001 bt002 bt003 bt004 bt010 bt010 bt011 bt012 bt014 bt015 bt016
bt017 bt018

I've a question if the grammars shouldn't be done as
  blastp_2_0_10
  blastn_2_0_10
   ...
then define blast_2_0_10 = blastp_2_0_10 | blastn_2_0_10 | ...

"""

from Martel import *

blastp_version = Group("version",
                       Re(r"(?P<program>BLASTP) 2\.0\.10 \[Aug-26-1999\]")
                       ) + AnyEol()
blastn_version = Group("version",
                       Re(r"(?P<program>BLASTN) 2\.0\.10 \[Aug-26-1999\]")
                       ) + AnyEol()

blastx_version = Group("version",
                       Re(r"(?P<program>BLASTX) 2\.0\.10 \[Aug-26-1999\]")
                       ) + AnyEol()

tblastn_version = Group("version",
                       Re(r"(?P<program>TBLASTN) 2\.0\.10 \[Aug-26-1999\]")
                       ) + AnyEol()

tblastx_version = Group("version",
                       Re(r"(?P<program>TBLASTX) 2\.0\.10 \[Aug-26-1999\]")
                       ) + AnyEol()

version = blastp_version | blastn_version | blastx_version | \
          tblastn_version | tblastx_version

blank_line =  Group("blank", Re(" *\n"))

# Must have at least one character
line = Group("line", Re("[^\R]+\R"))

reference = Group("reference",
                  line + line + line + line)


query_info = Str("Query= ") + ToEol("query-def") + \
             Rep(Re(r"(?P<query-def>(?!   )[^\R]*)\R")) + \
             Re(r"\s+\((?P<query-len>\d+) letters\)\R")

database = Str("Database: ") + ToEol("db") + \
           Re(r"\s+(?P<num_sequences>[\d,]+) sequences; " \
              r"(?P<num_letters>[\d,]+) total letters\R")

#####

index_header = Str("""\
                                                                   Score     E
Sequences producing significant alignments:                        (bits)  Value
""")

index_line = Re(r"(?P<query-def>.{66})\s*(?P<score>\d+)\s+(?P<e-value>[^\R]+)\R")

index = index_header + blank_line + Rep(index_line)



hit_desc = Re(r"(?!           Length = \d+\R)(?P<desc>.*)\R")
hit_description = Str(">") + Group("hit_description",
                                   Rep(hit_desc))
hit_length = Str("           Length = ") + \
             Group("hit-len", Re("\d+")) + \
             AnyEol()

hit1 = Re(r" Score =\s*(?P<score>[\d.]+) bits \((?P<bits>\d+)\)," \
          r" Expect(\((?P<expect_num>\d)\))? = (?P<expect>[^\R]+)\R")
hit2 = Re(r" Identities = (?P<num_identities>\d+)/\d+ \(\d+%\)") + \
       Opt(Re(r", Positives = (?P<num_positives>\d+)/\d+ \(\d+%\)")) + \
       Opt(Re(r", Gaps = (?P<num_gaps>\d+)/\d+ \(\d+%\)")) + \
       AnyEol()

#XX add test!
hit3a = Re(" Strand = (?P<query_strand>(Plus|Minus)) / " + \
          "(?P<subject_strand>(Plus|Minus))\R\R")

#XX add test!
hit3b = Re(" Frame = (?![+-]\d+ /)(?P<frame>[+-]\d+)\R")

#XX add test!
hit3c = Re(" Frame = (?P<query_frame>[+-]\d+) / (?P<subject_frame>[+-]\d+)\R")

hit_header = hit_description + hit_length + Re(r" *\R")

subhit_header = hit1 + hit2 + Opt(hit3a | hit3b | hit3c)


query_line = Group("query_line_header",
                   Re("Query: (?P<query_start>\d+)\s*")) + \
             Group("query_hit_seq", Re("\S*")) + \
             Re(r"\s+(?P<query_end>\d+)\R")
midline = ToEol("midline")
sbjct_line = Group("subject_line_header",
                   Re("Sbjct: (?P<subject_start>\d+)\s*")) + \
             Group("subject_hit_seq", Re("\S*")) + \
             Re(r"\s+(?P<query_end>\d+)\R")

hit_alignment = Rep(blank_line) + query_line + midline + sbjct_line

hit = hit_header + \
      Rep1(Group("subhit", subhit_header +
                 Group("hit_alignment",
                       Rep1(hit_alignment)) + blank_line + blank_line))

statistics1 = Re("""\
  Database: (?P<database>.*)
    Posted date:  (?P<date>.*)
  Number of letters in database: (?P<num_letters>[\\d,]+)
  Number of sequences in database:  (?P<num_sequences>[\\d,]+)
 *
""", fix_newlines = 1)

statistics2 = Re(r"""Lambda     K      H
\s*(?P<lambda1>[\d.]+)\s*(?P<k1>[\d.]+)\s+(?P<h1>[\d.]+) *

(((?P<gapped>Gapped)\R)?Lambda     K      H
\s*(?P<lambda2>[\d.]+)\s*(?P<k2>[\d.]+)\s+(?P<h2>[\d.]+) *

)?
""", fix_newlines = 1)
statistics3 = Re(r"""Matrix: (?P<matrix>.+)
(Gap Penalties: Existence: (?P<existence_penalty>\d+), Extension: (?P<extension_penalty>\d+)
)?Number of Hits to DB: (?P<num_hits_to_db>[\d,]+)
Number of Sequences: (?P<num_sequences>[\d,]+)
Number of extensions: (?P<num_extensions>[\d,]+)
Number of successful extensions: (?P<num_successful_sequences>[\d,]+)
Number of sequences better than 10.0: (?P<num_better>[\d,]+)
(Number of HSP's better than 10.0 without gapping: (?P<num_better_nogap>[\d,]+)
Number of HSP's successfully gapped in prelim test: (?P<num_success_gap>[\d,]+)
Number of HSP's that attempted gapping in prelim test: (?P<num_attempted_gap>[\d,]+)
Number of HSP's gapped \(non-prelim\): (?P<num_gapped>[\d,]+)
)?""", fix_newlines = 1)
statistics4 = Re(r"""length of query: (?P<query_length>[\d,]+)
length of database: (?P<database_length>[\d,]+)
effective HSP length: (?P<effective_hsp_length>[\d,]+)
effective length of query: (?P<effective_query_length>[\d,]+)
effective length of database: (?P<effective_database_length>[\d,]+)
effective search space: (?P<effective_search_space>[\d,]+)
effective search space used: (?P<effective_search_used>[\d,]+)
(frameshift window, decay const: (?P<frameshift_window>\d+),  (?P<frameshift_decay_const>[\d.]+)
)?T: (?P<T>\d+)
A: (?P<A>\d+)
X1: (?P<X1>\d+) \((?P<X1_bits>\s*[\d.]+) bits\)
X2: (?P<X2>\d+) \((?P<X2_bits>\s*[\d.]+) bits\)
(X3: (?P<X3>\d+) \((?P<X3_bits>\s*[\d.]+) bits\)
)?S1: (?P<S1>\d+) \((?P<S1_bits>\s*[\d.]+) bits\)
S2: (?P<S2>\d+) \((?P<S2_bits>\s*[\d.]+) bits\)
""", fix_newlines = 1)

statistics = statistics1 + statistics2 + statistics3 + statistics4

format = Group("blast:2.0.10",
               version + \
               blank_line + \
               blank_line + \
               reference + \
               blank_line + \
               query_info + \
               blank_line + \
               database + \
               blank_line + \
               Rep(line) + \
               blank_line + \
               Alt(index + blank_line + Rep(hit),
                   Str(" ***** No hits found ******") + AnyEol() + blank_line,
                   Rep(hit),
                   )+ \
               statistics
               )

