"""Format from EMBL Nucleotide Sequence Database Release 65, December 2000

"""
import Martel

import swissprot38

def UpTo(c, name):
    return Martel.Group(name, Martel.Re(r"[^%s\R]+" % c))

whitespace = Martel.Re(" +")

## ID - identification             (begins each entry; 1 per entry)
# ID   entryname  dataclass; molecule; division; sequencelength BP.

divisions = Martel.Re("EST|PHG|FUN|GSS|HTC|HTG|HUM|INV|ORG|MAM|VRT|PLN|" + \
                      "PRO|ROD|SYN|STS|UNC|VRL|[A-Z]{3}")

# XXX is found in S40706
ID_line = Martel.Str("ID   ") + \
          UpTo(" ", "entryname") + \
          whitespace + \
          UpTo(";", "dataclass") + \
          Martel.Str("; ") + \
          Martel.Group("molecule",
                       Martel.Re("DNA|RNA|circular DNA|circular RNA|XXX")) + \
          Martel.Str("; ") + \
          Martel.Group("division", divisions) + \
          Martel.Str("; ") + \
          Martel.Integer("length") + \
          Martel.Str(" BP.") + \
          Martel.AnyEol()


## AC - accession number           (>=1 per entry)
accession = UpTo(";", "accession") + Martel.Str(";")
AC_line = Martel.Str("AC   ") + \
          accession + Martel.Rep(Martel.Str(" ") + accession) + \
          Martel.AnyEol()
          
AC_block = Martel.Rep1(AC_line)

## SV - sequence version           (1 per entry)
SV_line = Martel.Str("SV   ") + \
          Martel.Group("sequence_version",
                       UpTo(".", "accession") + \
                       Martel.Str(".") + \
                       Martel.Integer("version")) + \
          Martel.AnyEol()


## DT - date                       (2 per entry)
date = Martel.Integer("day") + \
       Martel.Str("-") + \
       UpTo("-", "month") + \
       Martel.Str("-") + \
       Martel.Integer("year")

DT_created_line = Martel.Str("DT   ") + \
                  Martel.Group("date_created", date) + \
                  Martel.Str(" (Rel. ") + \
                  Martel.Integer("release_created") + \
                  Martel.Str(", Created)") + \
                  Martel.AnyEol()

DT_updated_line = Martel.Str("DT   ") + \
                  Martel.Group("date_updated", date) + \
                  Martel.Str(" (Rel. ") + \
                  Martel.Integer("release_updated") + \
                  Martel.Str(", Last updated, Version ") + \
                  Martel.Integer("version_number") + \
                  Martel.Str(")") + \
                  Martel.AnyEol()

DT_block = DT_created_line + DT_updated_line

## DE - description                (>=1 per entry)
DE_line = Martel.Str("DE   ") + \
          Martel.ToEol("description")
DE_block = Martel.Group("description_block",
                        Martel.Rep1(DE_line))

## KW - keyword                    (>=1 per entry)
KW_line = Martel.Str("KW   ") + \
          Martel.ToEol("keyword_data")
KW_block = Martel.Rep1(KW_line)

## OS - organism species           (>=1 per entry)
OS_block = swissprot38.OS_block

## OC - organism classification    (>=1 per entry)
OC_block = swissprot38.OC_block

## OG - organelle                  (0 or 1 per entry)
OG_block = swissprot38.OG_block

organism = Martel.Group("organism",
                        OS_block + \
                        OC_block + \
                        Martel.Opt(OG_block))

## RN - reference number           (>=1 per entry)
## RC - reference comment          (>=0 per entry)
## RP - reference positions        (>=1 per entry)
## RX - reference cross-reference  (>=0 per entry)
## RA - reference author(s)        (>=1 per entry)
## RT - reference title            (>=1 per entry)
## RL - reference location         (>=1 per entry)
RN_line = swissprot38.RN
RC_block = swissprot38.RC_block
RP_line = swissprot38.RP

RX_line = swissprot38.RX
RX_block = Martel.Group("RX_block", Martel.Rep1(RX_line))

RA_block = swissprot38.RA_block
RT_block = swissprot38.RT_block
RL_block = swissprot38.RL_block

reference = Martel.Group("reference",
                         RN_line + \
                         Martel.Opt(RC_block) + \
                         Martel.Opt(RP_line) + \
                         Martel.Opt(RX_block) + \
                         RA_block + \
                         RT_block + \
                         RL_block)

## DR - database cross-reference   (>=0 per entry)
DR_block = swissprot38.DR_block

## FH - feature table header       (0 or 2 per entry)
FH_block = Martel.Str("FH   Key             Location/Qualifiers") + \
           Martel.AnyEol() + \
           Martel.Str("FH") + \
           Martel.AnyEol()

## FT - feature table data         (>=0 per entry)
FT_line = Martel.Str("FT   ") + \
          Martel.ToEol("ft_data")
FT_block = Martel.Rep1(FT_line)

## CC - comments or notes          (>=0 per entry)
CC_line = Martel.Str("CC   ") + \
          Martel.ToEol("comment")
CC_block = Martel.Rep1(CC_line)

## XX - spacer line                (many per entry)
XX = Martel.Str("XX") + Martel.AnyEol()

## SQ - sequence header            (1 per entry)
SQ_line = Martel.Str("SQ   Sequence ") + \
          Martel.Integer("num_BP") + \
          Martel.Str(" BP; ") + \
          Martel.Integer("num_A") + \
          Martel.Str(" A; ") + \
          Martel.Integer("num_C") + \
          Martel.Str(" C; ") + \
          Martel.Integer("num_G") + \
          Martel.Str(" G; ") + \
          Martel.Integer("num_T") + \
          Martel.Str(" T; ") + \
          Martel.Integer("num_other") + \
          Martel.Str(" other;") + \
          Martel.AnyEol()

## bb - (blanks) sequence data     (>=1 per entry)
SQ_data = Martel.Str("     ") + \
          Martel.Group("sequence", Martel.Re(".{65}")) + \
          whitespace + \
          Martel.Integer("end_position") + \
          Martel.AnyEol()

SQ_block = SQ_line + Martel.Rep1(SQ_data)

## // - termination line           (ends each entry; 1 per entry)
end = Martel.Str("//") + Martel.AnyEol()

record = Martel.Group("embl65_record", \
                      ID_line + \
                      Martel.Opt(XX) + \
                      AC_block + \
                      Martel.Opt(XX) + \
                      SV_line + \
                      Martel.Opt(XX) + \
                      DT_block + \
                      Martel.Opt(XX) + \
                      DE_block + \
                      Martel.Opt(XX) + \
                      KW_block + \
                      Martel.Opt(XX) + \
                      Martel.Rep1(organism + Martel.Opt(XX)) + \
                      Martel.Rep(reference + Martel.Opt(XX)) + \
                      Martel.Opt(DR_block + \
                                 Martel.Opt(XX)) + \
                      Martel.Rep(CC_block + \
                                 Martel.Opt(XX)) + \
                      FH_block + \
                      FT_block + \
                      Martel.Opt(XX) + \
                      SQ_block + \
                      end)

format = Martel.Rep1(record)
