"""Read the PCR primers database from http://www.ebi.ac.uk/primers_home.html

I don't know which fields are required/optional nor do I know their
order.  What you see below is a very restrictive definition based on
what I found from the primers.dat file but allowing little other
variations.
"""

import Martel
from Martel import RecordReader

def Simple(name, tag):
    assert len(name) == 6, (name, len(name))
    return Martel.Str(name + "  ") + Martel.ToEol(tag)

def Block(name, tag):
    return Martel.Group(tag + "_block",
                        Martel.Rep1(Martel.Str(name + "  ") + Martel.ToEol(tag)))

def Line(name, tag, exp):
    return Martel.Str(name + "  ") + Martel.Group(tag, exp) + Martel.AnyEol()

## ACCESS  Accession number. Composed of a constant two characters PR
##         followed by 5 alphanumeric characters defaulting to numeric as long
##         as the address space is sufficient. Appears only once.
# ACCESS  PR00001
access = Simple("ACCESS", "accession_number")

## XXXXXX  A blank separator line. May repeat more then once.
# XXXXXX
sep = Martel.Rep1(Martel.Str("XXXXXX") + Martel.Re(" *") + Martel.AnyEol())
 
## DATEIN  Submission / updates date   Followed by either (created) or
##         (update). May appear more than once.
# DATEIN  9-NOV-1995  (created)
date = Martel.Re(r"(?P<day>\d+)-(?P<month>\w+)-(?P<year>\d+)")
datein_created = Line("DATEIN",
                      "datein_created",
                      Martel.Group("date_created", date) + \
                      Martel.Re(r" *\(created\)"))
datein_update = Line("DATEIN",
                     "datein_update",
                     Martel.Group("date_update", date) + \
                     Martel.Re(r" *\(update\)"))

## TARGET  The amplification target. This may be a specification of a known
##         gene name, a chromosome and possibly the location on it etc.
##         Specifying the organism here is optional. It may be multi line up
##         to 250 characters.
# TARGET  miniswine MHC class I gene - PD1 expressed in transgenic mice
#
# TARGET  Mouse MHC class I gene - H-2Kd, derived from clone pH2D33,
# TARGET  expressed in transgenic mice
target = Block("TARGET", "target")

## NUMREF  Specification of what do the sequence numbering and positions
##         refers to. It may either be "Position 1 on cDNA", "A database
##         entry" (in which case a cross reference should be provided in the
##         DBXREF field), "Start codon of a gene", or where needed should be
##         defined by the submitting author following this outline of
##         definition.
# NUMREF  position 1 on a cDNA
# NUMREF  EMBL/Genbank/DDBJ:D28581
# NUMREF  Y00052 M19533 X52803 X17105
# NUMREF  Position 1 on 16S RNA to position 742 on 23S RNA
numref = Block("NUMREF", "numref")

## OSPECS  Organism Species. Defined following the EMBL/Genbank/DDBJ
##         definitions
# OSPECS  SUS SCROFA (PIG)
ospecs = Simple("OSPECS", "ospecs")

## SPAPLL  Special applications. Free text with an aim for establishing a
##         standard nomenclature.
# SPAPLL  Amplifies human target in human-rodent somatic cell hybrids
spapll = Simple("SPAPLL", "spapll")

## LNCDNA  Product length in base pairs when a cDNA is amplified.
# LNCDNA  217
# LNCDNA  variable (87bp without insert)
lncdna = Simple("LNCDNA", "lncdna")
 
## LNGDNA  Product length in base pairs when a Genomic DNA is amplified.
# LNGDNA  122
# LNGDNA  230 bp
# LNGDNA  ~500bp (12s); ~800 bp (16s); ~1200 (18s)
# LNGDNA  5.75 Kb
# LNGDNA  340  bp
lngdna = Simple("LNGDNA", "lngdna")
 
## FPNAME  Name of the forward (5' --> 3') primer, assigned by the submitting
##         author.
# FPNAME  PD1X2P132
# FPNAME  D30.31
# FPNAME  Ac1382Forward
# FPNAME  primer_1
fpname = Simple("FPNAME", "fpname")

## FPSEQN  Forward primer sequence, provided in 5' --> 3' direction (i.e.
##         ready for synthesis). Sequence therefore will be given only with
##         non-ambiguous bases. It will be displayed in triades, uppercase,
##         followed by the sequence length.
# FPSEQN  TAT CAC CCC AGA TCT CAA GTT GAT     24
# FPSEQN  CTG AAG CTG CCA ATT CC     17
fpseqn = Line("FPSEQN", "fpseqn",
                Martel.Re("(?P<bases>((?!  )[ATCG ])+) *(?P<primer_size>\d+)"))

## FFLANK  Flanking base on the target sequence. i.e. Which base on the target
##         sequence matches the first 5' base of the primer. This position
##         will be given with reference to the NUMREF field given earlier.
# FFLANK  643
# FFLANK  -45
# FFLANK  218 (Y00118)
fflank = Simple("FFLANK", "fflank")

## FRESTR  Restriction site added to the primer sequence. The names of the
##         restriction enzymes will follow the nomenclature of the REBASE
##         database.
# FRESTR  BamH
frestr = Simple("FRESTR", "frestr")

## FMUTAT  Mutation introduced into the primer sequence. Will be expressed as
##         base1 (or bases1) --> base2 (or bases2) in position, whereas base1
##         is the  base(s) in the original sequence, base2 is the base(s) that
##         were  introduced as a mutation and position is the position on the
##         primer.
# Cannot find any examples
fmutat = Simple("FMUTAT", "fmutat")

## RPNAME  Name of the reverse (3' --> 5') primer, assigned by the submitting
##         author.
# RPNAME  H2KdX5AS1020
# RPNAME  D30.16
# RPNAME  primer_2
rpname = Simple("RPNAME", "rpname")

## RPSEQN  Reverse primer sequence, provided in 5' --> 3' direction (i.e.
##         ready for synthesis). Sequence therefore will be given only with
##         non-ambiguous bases. It will be displayed in triades.
# RPSEQN  GGA GGA TAG AGA CAC AC     17
# RPSEQN  ATG TTC TTC TTG GGC CTT ATC TAT TCC A     28
rpseqn = Line("RPSEQN", "rpseqn",
                Martel.Re("(?P<bases>((?!  )[ATCG ])+) *(?P<primer_size>\d+)"))

## RFLANK  See definition for FFLANK.
# RFLANK  7
# RFLANK  291 (Y00118); 41 (J02460)
# RFLANK  GGAAGGGCATTAAGGTCATTA
rflank = Simple("RFLANK", "rflank")

## RRESTR  See definition for FRESTR.
# RRESTR  Kpn I
# RRESTR  Mlu
rrestr = Simple("RRESTR", "rrestr")

## RMUTAT  See definition for FMUTAT.
# Cannot find any examples
rmutat = Simple("RMUTAT", "rmutat")

## TCYCLE  Thermal Cycler make and model. Two fields separated by semicolon
##         make;model
# TCYCLE  MJ Research; minicycler
# TCYCLE  Perkin-Elmer; DNA thermal cycler
# TCYCLE  Applied Biosystems; GeneAmp PCR system 9600
tcycle = Line("TCYCLE", "tcycle",
              Martel.Re(r"(?P<cycler_make>[^;\R]+); "
                        r"(?P<cycler_model>[^;\R]+)"))

## PROBIN  Temperature Sampling By Probe In-tube. A Boolean field accepting
##         either "YES" or "NO"
# PROBIN  YES
# PROBIN  NO
probin = Line("PROBIN", "probin",
              Martel.Alt(Martel.Str("YES"), Martel.Str("NO")))

## HOTSTR  Was the Hot-start protocol applied? A Boolean field accepting
##         either "YES" or "NO"
# HOTSTR  NO
# HOTSTR  YES
hotstr = Line("HOTSTR", "hotstr",
              Martel.Alt(Martel.Str("YES"), Martel.Str("NO")))

## IDENAT  Initial denaturation. Expressed as two fields separated by
##         semicolon. Temperature in degrees centigrade ; duration in seconds.
# IDENAT  92; 300
# IDENAT  95; 30
idenat = Line("IDENAT", "idenat",
              Martel.Integer("temperature") + \
              Martel.Str("; ") + \
              Martel.Integer("duration"))

## DENATR  Denaturation. Expressed as two fields separated by semicolon.
##         Temperature in degrees centigrade ; duration in seconds.
# DENATR  94; 45
denatr = Line("DENATR", "denatr",
              Martel.Integer("temperature") + \
              Martel.Str("; ") + \
              Martel.Integer("duration"))

## ANNEAL  Annealing.  Expressed as two fields separated by semicolon.
##         Temperature in degrees centigrade ; duration in seconds.
# ANNEAL  57; 45
# ANNEAL  55; 160
anneal = Line("ANNEAL", "anneal",
              Martel.Integer("temperature") + \
              Martel.Str("; ") + \
              Martel.Integer("duration"))

## TCHDWN  Touchdown protocol. Has two fields, separated by semicolon:
##         Temperature decrease in degrees centigrade; Number of cycles in
##         which touchdown was applied.
# TCHDWN  0.4; 2
tchdwn = Line("TCHDWN", "tchdwn",
              Martel.Float("temperature_decrease") + \
              Martel.Str("; ") + \
              Martel.Integer("cycle_number"))

## EXTEND  Extension.  Expressed as two fields separated by semicolon.
##         Temperature in degrees centigrade ; duration in seconds.
# EXTEND  72; 30
# EXTEND  72; 120
extend = Line("EXTEND", "extend",
              Martel.Integer("temperature") + \
              Martel.Str("; ") + \
              Martel.Integer("duration"))

## PRGRSS  Progressive extension. Has two fields, separated by semicolon: The
##         time added in seconds per cycle; The cycle number in which addition
##         has started.
# Cannot find any examples
prgrss = Simple("PRGRSS", "prgrss")

## FINEXT  Final extension step. Expressed as two fields separated by
##         semicolon. Temperature in degrees centigrade ; duration in seconds.
# FINEXT  72; 10
# FINEXT  72; 300
finext = Line("FINEXT", "finext",
              Martel.Integer("temperature") + \
              Martel.Str("; ") + \
              Martel.Integer("duration"))

## CYCLES  Total number of cycles per amplification reaction.
# CYCLES  35
cycles = Simple("CYCLES", "cycles")

## RXNVOL  Total reaction volume expressed in microlitres.
# RXNVOL  25
rxnvol = Simple("RXNVOL", "rxnvol")

## POLYMR  Polymerase used. Expressed as three fields separated by semicolons:
##         Polymerase type; Polymerase make; Polymerase concentration in Units
##         per reaction.
# POLYMR  Taq; Apligene; 0.5
# POLYMR  Taq; Perkin Elmer AmpliTaq; 2.5
# POLYMR  Pfu; Stratagene; 2.5
# POLYMR  Taq; Pharmacia; 1
# POLYMR  Taq; Tth+; 0.3
polymr = Martel.Group("polymr_block",
                      Martel.Rep1(
    Line("POLYMR",
         "polymr",
         Martel.Re("(?P<polymerase_type>[^;\R]+);") + \
         Martel.Re(" *(?P<polymerase_make>[^;\R]+);") + \
         Martel.Re(" *(?P<polumerase_concentration>[^\R]+"))
    ))

## BUFFER  The type of buffer used. Can be either a commercially standard
##         provided, in which case, the name of the provider is stated, or it
##         can be one of a few standard buffers of the database, or provided
##         in full details.
# BUFFER  Promega
# BUFFER  10 x AmpliTaq (Perkin Elmer)
# 
# BUFFER  The standard Taq (200 mM Tris-HCL pH 8.4, 500 mM KCL) polymerase
# BUFFER  buffer from GibcoBRL was used at a 1X concentration
buffer = Block("BUFFER", "buffer")

## DNTPCN  Concentration of dNTP's in uM (micro molar). If all are the same,
##         only one figure will be provided. Every dNTP with a different
##         concentration follows separated by a semicolon.
# DNTPCN  400
# DNTPCN  0.1
# Cannot find examples with different concentrations
dntpcn = Simple("DNTPCN", "dntpcn")

## PRIMCN  Concentration of primers in uM (micro molar). If both are the same,
##         only one figure will be provided. Else, the concentration of each
##         is provided followed by the primer's name as appeared in FPNAME and
##         RPNAME separated by a semicolon.
# PRIMCN  2
# PRIMCN  1.5
# Cannot find examples where the concentrations differ
primcn = Simple("PRIMCN", "primcn")

## MGCONC  Concentration of MgCl2 in the buffer provided in mM (millimolar).
##         Field will be included only if the  MgCl2 concentration varied from
##         a standard buffer as defined in the BUFFER field. The figure
##         provided will allways refer to the final total concentration of
##         Magnesium in the buffer.
# MGCONC  1.5
# MGCONC  20
mgconc = Simple("MGCONC", "mgconc")

## ADITIV  Various additives added to the reaction (e.g. DMSO, Formamide,
##         Gelatin ). The units will be provided according to the submitters
##         specifications. In the field the order will allways be additive
##         name, concentration, units and be separated by a semicolon;
# ADITIV  BSA 1%; DMSO 5%
# ADITIV  Cosmid;
additive = Martel.Re("(?P<additive>[^;\R]+);?")
aditiv = Line("ADITIV", "aditiv",
              additive + Martel.Rep(Martel.Str(" ") + additive))

## SOURCE  DNA Source, Method of Preparation and Amount added.
# SOURCE  5ul of blood + 1ml of DDW + 10% Chelex X-100 (BioRad). Boil 20
# SOURCE  minutes. Take 5ul of supernatant / reaction
#
# SOURCE  Cosmid Nucleobond preparation; 10ng
#
# SOURCE  Germinating embryos, Doyle and Doyle (1990), 25ng
source = Block("SOURCE", "source")

## DBXREF  Cross references for other databases. Expressed as:
##         Database name  followed by a colon and the the acession numbers
##         separated by a comma. Various  databases separated with semicolons.
##         Database names according to  pre-defined nomenclature.
# DBXREF  EMBL/Genbank/DDBJ: X95489; X95490; X95491; X95492; X95493
# DBXREF  EMBL/Genbank/DDBJ:M21057; M19649;
# DBXREF  EMBL/Genbank/DDBJ:S67722
# DBXREF  EMBL/Genbank/DDBJ:X95404;
# DBXREF  EMBL/Genbank/DDBJ:Y00052; M19533; X52803; X17105;
# DBXREF  PIR:A02204;
# DBXREF  SWISS-PROT: P43118;
# DBXREF  SWISS-PROT:P01902;
#   XXX *you* write the regexp for this nonsense :)
dbxref = Block("DBXREF", "dbxref")

## CNTNAM  Names of contact authors in the following format: First name
##         Surname, separated by a comma.
# CNTNAM  Marc Lipinski
# CNTNAM  MARC LIPINSKI
# CNTNAM  Clemens Suter-Crazzolara
# CNTNAM  Edmundo Castro (submits) Dr. Gorden Glazner (designer)
# CNTNAM  motooki ashihara
# CNTNAM  dennis S SALONGA
cntnam = Simple("CNTNAM", "cntnam")

## CNTINS  Institute of contact authors
# CNTINS  EMBL outstation, the EBI, User support
# CNTINS  Forest tree breeding institute, National research council
# CNTINS  Forest tree breeding institute, National research council
# CNTINS  helix research institute, the third division
cntins = Simple("CNTINS", "cntins")

## CNTADR  Regular mail address of contact authors
# CNTADR  5929 Life Science Building, 405 Hilgard Ave., LA Ca 90095, USA
# CNTADR  5929 Life Science Building, 405 Hilgard Ave WLA, Ca 90095, USA
#
# CNTADR  5929 Life Sciences Building, University of California, Los
# CNTADR  Angeles, CA, 90095, USA
#
cntadr = Block("CNTADR", "cntadr")

## CNTEML  Email address of contact authors
# CNTEML  no_uce@here.there
cnteml = Simple("CNTEML", "cnteml")

## CNTTEL  Telephone number of contact authors, including international dial
##         code
# CNTTEL  310-2064123
# CNTTEL  +33 1 4559 49 23
# CNTTEL  +49-211-8114923
cnttel = Simple("CNTTEL", "cnttel")

## CNTFAX  Fax number of contact authors, including international dial code
# CNTFAX  +27-12-808-1423
# CNTFAX  310-206-9123
# CNTFAX  310-2069123
# CNTFAX  +33 1 4559 64 23
cntfax = Simple("CNTFAX", "cntfax")

## REFAUT  Reference authors in the following format: Surname, Initials.,
# REFAUT  Lorain S., Demczuk S., Lamour V., Toth S., Aurias A., Roe B.A.,
# REFAUT  Lipinski M.
#
# REFAUT  Adams JW., Migita DS., Yu MK., Young R., Hellickson MS.,
# REFAUT  Castro-Vargas FE., Domingo JD., Lee PH., Bui JS., Henderson SA.
refaut = Block("REFAUT", "refaut")

## REFTTL  Reference title
# REFTTL  Prostaglandin F2 alpha Stimulates Hypertrophic Growth of Cultured
# REFTTL  Neonatal Rat Ventricular Myocytes
#
# REFTTL  (Unpublished)
refttl = Block("REFTTL", "refttl")

## REFSRC  Reference publication line in the following format: Journal name
##         according to the nomenclature defined for EMBL database followed by
##         Volume: frompage - topage, (Year)
# REFSRC  Tel-Aviv University, Ph.D. thesis (1995)
# REFSRC  Genome Research
# REFSRC  J. Biol. Chem. 271:1179-1186 (1996)
# REFSRC  AIDS 7 (suppl 2):59-64 (1993)
refsrc = Simple("REFSRC", "refsrc")

## REFXRF  Reference cross-reference  (Typically a link to the appropriate
##         accession number in the public version of Medline available through
##         NCBI)
# REFXRF  MEDLINE; 95210610.
refxrf = Simple("REFXRF", "refxrf")

## REMARK  The name says it all... :-)  Anything that doesn't go into any of
##         the above fields, but needs to go in.
# REMARK  NOTE: Base number 7 in primer_2 has been designed to mismatch the
# REMARK  original sequence in K03455. This reflects the more frequent
# REMARK  appearance of T in the consensus 1.
#
# REMARK  Internal BamHI site is present in target sequence
remark = Block("REMARK", "remark")

## //      The good old faithful record terminator
end = Martel.Str("//") + Martel.AnyEol()

record = access + \
         sep + \
         datein_created + \
         sep + \
         target + \
         sep + \
         Martel.Opt(numref + \
                    sep) + \
         ospecs + \
         sep + \
         Martel.Opt(spapll + \
                    sep) + \
         Martel.Opt(((lncdna + Martel.Opt(lngdna)) | lngdna) + \
                    sep) + \
         fpname + \
         fpseqn + \
         Martel.Opt(fflank) + \
         Martel.Opt(frestr) + \
         sep + \
         rpname + \
         rpseqn + \
         Martel.Opt(rflank) + \
         Martel.Opt(rrestr) + \
         sep + \
         tcycle + \
         probin + \
         hotstr + \
         sep + \
         idenat + \
         denatr + \
         anneal + \
         Martel.Opt(tchdwn) + \
         extend + \
         Martel.Opt(finext) + \
         cycles + \
         sep + \
         rxnvol + \
         polymr + \
         buffer + \
         dntpcn + \
         primcn + \
         Martel.Opt(aditiv) + \
         Martel.Opt(mgconc) + \
         source + \
         sep + \
         Martel.Opt(dbxref + \
                    sep) + \
         cntnam + \
         cntins + \
         cntadr + \
         cnteml + \
         Martel.Opt(cnttel) + \
         Martel.Opt(cntfax) + \
         sep + \
         refaut + \
         Martel.Opt(refttl) + \
         Martel.Opt(refsrc) + \
         Martel.Opt(refxrf) + \
         Martel.Opt(sep) + \
         Martel.Opt(remark) + \
         end

         
format = Martel.ParseRecords("primers", record,
                             RecordReader.EndsWith, ("//\n",) )
