"""Parse the Prosite documentation file.  Tested against release 16.

This parser does extra work to identify [footnote,records], <PRODOC>
and (EC) links, which has about an extra 50% impact on performance (80
seconds instead of 50).

"""

import Martel
from Martel import RecordReader

prodoc_num = Martel.Re("PDOC\d{5}")
prosite_num = Martel.Re("PS\d{5}")

reference_note = Martel.Group("reference_note",
                 Martel.Re("\[(?P<note>E?\d+)(,(?P<note>E?\d+))*\]"))
prodoc_link = Martel.Group("prodoc_link",
                           Martel.Str("<") + prodoc_num + Martel.Str(">"))
# This is incomplete and doesn't allow things like 1.1.-.-
ec_link = Martel.Group("ec_link",
    Martel.Re("\(EC *(?P<ec_number>[1-9][0-9]*\.[1-9][0-9]*\.[1-9][0-9]*\.([1-9][0-9]*|-))\)"))
                       
generic_text = reference_note | \
               prodoc_link | \
               ec_link | \
               Martel.Re("[^\R]")

# {PDOC00012}
prodoc_start = Martel.Str("{") + \
               Martel.Group("prodoc_id", prodoc_num) + \
               Martel.Str("}") + \
               Martel.AnyEol()

# {PS50075; ACP_DOMAIN}
prosite_ref = Martel.Str("{") + \
              Martel.Group("prosite_id", prosite_num) + \
              Martel.Str("; ") + \
              Martel.Group("prosite_acc", Martel.Re(r"\w+")) + \
              Martel.Str("}") + \
              Martel.AnyEol()

begin = Martel.Str("{BEGIN}") + Martel.AnyEol()

# The title is the part inside of "***"s.  It can fold over several lines.
title = Martel.Group("title",
                     Martel.Re(r"\*+\R") + \
                     Martel.Rep1(
    Martel.Re(r"\* (?P<title_text>((?! \*)[^\R])+) \*\R")) + \
                     Martel.Re(r"\*+\R"))

text = Martel.Group("text", Martel.Rep(
    Martel.Re("(?!-)") + \
    Martel.Rep(generic_text) + \
    Martel.AnyEol()))

# All sorts of different ways to spell the same thing, like
#   -Other sequence(s)  detected  in  SWISS-PROT
#   -Other sequence(s) detected in SWISS-PROT
# Blech, and there's a "o-" in PDOC00573

update_info = Martel.Group("update_information",
                           Martel.Str("-Last update: ") + \
                           Martel.Re(r"(?P<month>\w+) (?P<year>\d+) / ") + \
                           Martel.ToEol("update_status"))
expert_info = Martel.Group("expert_information",
                           Martel.Str("-Expert(s) to contact by email:") + \
                           Martel.AnyEol() + \
                           Martel.Rep1(
                                Martel.Str("           ") + \
                                Martel.Re(r"(?P<email_name>[^;\R]+); " \
                                          r"(?P<email_address>[^\R]+)\R")
                                )
                           )

generic_info = \
   ((Martel.Str("-") + Martel.Rep(generic_text) + Martel.AnyEol()) | \
    (Martel.Str("o-Sequences") + Martel.Rep(generic_text) + Martel.AnyEol())
    ) + \
   Martel.Rep(Martel.Str(" ") + Martel.Rep(generic_text) + Martel.AnyEol())

info = Martel.Group("information",
    (update_info |
     expert_info |
     generic_info) + \
    Martel.Rep(Martel.AnyEol()))

# Footnotes on the bottom like "[ 1]" or "[E2]"
url_footnote = Martel.Group("footnote",
                            Martel.Re(r"\[ *(?P<footnote_id>E\d+)\] ") + \
                            Martel.ToEol("url_link"))

journal_footnote = Martel.Group("footnote",
                                Martel.Re(r"\[ *(?P<footnote_id>\d+)\] ") + \
                                Martel.ToEol("footnote_text") + \
                                Martel.Rep(Martel.Str("    ") + \
                                           Martel.Re(" *") + \
                                           Martel.ToEol("footnote_text")))
footnote = journal_footnote | url_footnote
                                   

end = Martel.Str("{END}") + Martel.AnyEol()

## Special case PDOC00000
pdoc00000_start = Martel.Str("{PDOC00000}") + Martel.AnyEol()
pdoc00000_title = \
        Martel.Group("title",
          Martel.Str("**********************************") + Martel.AnyEol() + \
          Martel.Str("*** ") + \
          Martel.Group("title_text",
                       Martel.Str("PROSITE documentation file")) + \
          Martel.Str(" ***") + Martel.AnyEol() + \
          Martel.Str("**********************************") + Martel.AnyEol()
                     )

pdoc00000_text = Martel.Rep(Martel.Re(r"(?!\{END\})[^\R]*\R"))

pdoc00000_record = pdoc00000_start + \
                   begin + \
                   pdoc00000_title + \
                   pdoc00000_text + \
                   end

generic_record = prodoc_start + \
                 Martel.Rep1(prosite_ref) + \
                 begin + \
                 title + \
                 text + \
                 Martel.Opt(Martel.Rep1(info) + \
                            Martel.Rep(footnote)) + \
                 end

record = pdoc00000_record | \
         generic_record

format = Martel.ParseRecords("prodoc16", record,
                             RecordReader.EndsWith, ("{END}\n",))
